# Copyright 2015 Planet Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ._fatomic import atomic_open
from .exceptions import RequestCancelled
from .utils import get_filename
from .utils import check_status
from .utils import GeneratorAdapter
from datetime import datetime
import itertools
import json

chunk_size = 32 * 1024


class Response(object):

    def __init__(self, request, dispatcher):
        self.request = request
        self._dispatcher = dispatcher
        self._body = None
        self._future = None
        self._cancel = False

    def _create_body(self, response):
        return self.request.body_type(self.request, response, self._dispatcher)

    def get_body(self):
        '''Get the response Body

        :returns Body: A Body object containing the response.
        '''
        if self._body is None:
            resp = self._dispatcher._dispatch(self.request)
            self._body = self._create_body(resp)
        return self._body

    def _async_callback(self, session, response):
        if self._cancel:
            raise RequestCancelled()
        check_status(response)
        self._body = self._create_body(response)
        self._handler(self._body)
        if self._wait:
            self._wait(self._body)

    def get_body_async(self, handler, wait=None):
        if self._future is None:
            self._handler = handler
            self._wait = wait
            self._future = self._dispatcher._dispatch_async(
                self.request, self._async_callback
            )

    def wait(self):
        '''Await completion of this request.

        :returns Body: A Body object containing the response.
        '''
        if self._future:
            self._future.result()
        return self._body

    def cancel(self):
        '''Cancel any request.'''
        if self._body:
            self._body._cancel = True
        else:
            self._cancel = True


class Request(object):

    def __init__(self, url, auth, params=None, body_type=Response, data=None,
                 method='GET'):
        self.url = url
        self.auth = auth
        self.params = params
        self.body_type = body_type
        self.data = data
        self.method = method


class Body(object):
    '''A Body is a representation of a resource from the API.
    '''

    def __init__(self, request, http_response, dispatcher):
        self._request = request
        self.response = http_response
        self._dispatcher = dispatcher
        self.size = int(self.response.headers.get('content-length', 0))
        self._cancel = False

    @property
    def name(self):
        '''The name of this resource. The default is to use the
        content-disposition header value from the response.'''
        return get_filename(self.response)

    def __len__(self):
        return self.size

    def __iter__(self):
        return (c for c in self.response.iter_content(chunk_size=chunk_size))

    def last_modified(self):
        '''Read the last-modified header as a datetime, if present.'''
        lm = self.response.headers.get('last-modified', None)
        return datetime.strptime(lm, '%a, %d %b %Y %H:%M:%S GMT') if lm \
            else None

    def get_raw(self):
        '''Get the decoded text content from the response'''
        return self.response.content.decode('utf-8')

    def _write(self, fp, callback):
        total = 0
        if not callback:
            def noop(*a, **kw):
                pass
            callback = noop
        callback(start=self)
        for chunk in self:
            if self._cancel:
                raise RequestCancelled()
            fp.write(chunk)
            size = len(chunk)
            total += size
            callback(wrote=size, total=total)
        # seems some responses don't have a content-length header
        if self.size == 0:
            self.size = total
        callback(finish=self)

    def write(self, file=None, callback=None):
        '''Write the contents of the body to the optionally provided file and
        providing progress to the optional callback. The callback will be
        invoked 3 different ways:

        * First as ``callback(start=self)``
        * For each chunk of data written as
          ``callback(wrote=chunk_size_in_bytes, total=all_byte_cnt)``
        * Upon completion as ``callback(finish=self)``

        :param file: file name or file-like object
        :param callback: optional progress callback
        '''
        if not file:
            file = self.name
        if not file:
            raise ValueError('no file name provided or discovered in response')
        if hasattr(file, 'write'):
            self._write(file, callback)
        else:
            with atomic_open(file, 'wb') as fp:
                self._write(fp, callback)


class JSON(Body):
    '''A Body that contains JSON'''

    def get(self):
        '''Get the response as a JSON dict'''
        return self.response.json()


class Paged(JSON):

    ITEM_KEY = 'features'
    LINKS_KEY = '_links'
    NEXT_KEY = '_next'

    def next(self):
        links = self.get()[self.LINKS_KEY]
        next_ = links.get(self.NEXT_KEY, None)
        if next_:
            request = Request(next_, self._request.auth, body_type=type(self))
            return self._dispatcher.response(request).get_body()

    def _pages(self):
        page = self
        while page is not None:
            yield page
            page = page.next()

    def iter(self, pages=None):
        '''Get an iterator of pages.

        :param int pages: optional limit to number of pages
        :return: iter of this and subsequent pages
        '''
        i = self._pages()
        if pages is not None:
            i = itertools.islice(i, pages)
        return i

    def json_encode(self, out, limit=None, sort_keys=False, indent=None):
        '''Encode the results of this paged response as JSON writing to the
        provided file-like `out` object. This function will iteratively read
        as many pages as present, streaming the contents out as JSON.

        :param file-like out: an object with a `write` function
        :param int limit: optional maximum number of items to write
        :param bool sort_keys: if True, output keys sorted, default is False
        :param bool indent: if True, indent output, default is False
        '''
        stream = self._json_stream(limit)
        enc = json.JSONEncoder(indent=indent, sort_keys=sort_keys)
        for chunk in enc.iterencode(stream):
            out.write(u'%s' % chunk)

    def items_iter(self, limit):
        '''Get an iterator of the 'items' in each page. Instead of a feature
        collection from each page, the iterator yields the features.

        :param int limit: The number of 'items' to limit to.
        :return: iter of items in page
        '''

        pages = (page.get() for page in self._pages())
        items = itertools.chain.from_iterable(
            (p[self.ITEM_KEY] for p in pages)
        )
        if limit is not None:
            items = itertools.islice(items, limit)
        return items

    def _json_stream(self, limit):
        items = self.get()[self.ITEM_KEY]
        # if there are no results, the GeneratorAdapter doesn't play well
        if len(items):
            items = GeneratorAdapter(self.items_iter(limit))
        else:
            items = []
        return {
            self.ITEM_KEY: items
        }


# GeoJSON feature
class Features(Paged):

    def _json_stream(self, limit):
        stream = super(Features, self)._json_stream(limit)
        json_body = self.get()
        # patch back in the count if present
        if 'count' in json_body:
            stream['count'] = json_body.get('count')
        stream['type'] = 'FeatureCollection'
        return stream


class Items(Features):
    pass


class Searches(Paged):
    ITEM_KEY = 'searches'


class Mosaics(Paged):
    ITEM_KEY = 'mosaics'


class MosaicQuads(Paged):
    ITEM_KEY = 'items'


class AnalyticsPaged(Paged):
    LINKS_KEY = 'links'
    NEXT_KEY = 'next'
    ITEM_KEY = 'data'

    def next(self):
        links = self.get()[self.LINKS_KEY]
        next_ = None
        for link in links:
            if link['rel'] == self.NEXT_KEY:
                next_ = link['href']
        if next_:
            request = Request(next_, self._request.auth, body_type=type(self))
            return self._dispatcher.response(request).get_body()


# The analytics API returns two conceptual types of objects: WFS3-compliant
# objects and everything else. There may be some overlap (ex. subscriptions and
# collections).
class Feeds(AnalyticsPaged):
    pass


class Subscriptions(AnalyticsPaged):
    pass


class WFS3Paged(AnalyticsPaged):
    pass


class WFS3Collections(AnalyticsPaged):
    ITEM_KEY = 'collections'


class WFS3Features(AnalyticsPaged):
    # Explicitly disambiguate between WFS3 and GeoJSON features because the
    # differences in the structure of the response envelope result in paging
    # slightly differently.
    ITEM_KEY = 'features'


class Orders(Paged):
    ITEM_KEY = 'orders'


class Order(JSON):
    LINKS_KEY = '_links'
    RESULTS_KEY = 'results'
    LOCATION_KEY = 'location'

    def get_results(self):
        links = self.get()[self.LINKS_KEY]
        results = links.get(self.RESULTS_KEY, None)
        return results

    def get_locations(self):
        results = self.get_results()
        locations = [r[self.LOCATION_KEY] for r in results]
        return locations

    def items_iter(self, limit):
        '''Get an iterator of the 'items' in each order.
        The iterator yields the individual items in the order.

        :param int limit: The number of 'items' to limit to.
        :return: iter of items in page
        '''

        locations = iter(self.get_locations())

        # if limit is not None:
        #     locations = itertools.islice(locations, limit)
        return locations



# Copyright 2017 Planet Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re

import click
from click.types import CompositeParamType

from .util import read
from .util import _split

from planet.api import filters
from planet.api.utils import geometry_from_json
from planet.api.utils import strp_lenient
from planet.scripts.item_asset_types import get_item_types, get_asset_types, \
    get_bundles, DEFAULT_ITEM_TYPES, DEFAULT_ASSET_TYPES, DEFAULT_BUNDLES


metavar_docs = {
    'FIELD COMP VALUE...': '''A comparison query format where FIELD is a
    property of the item-type and COMP is one of lt, lte, gt, gte and VALUE is
    the number or date to compare against.

    Note: ISO-8601 variants are supported. For example, ``2017`` is short for
    ``2017-01-01T00:00:00+00:00``.
    ''',
    'FIELD VALUES...': '''Specifies an 'in' query where FIELD is a property
    of the item-type and VALUES is space or comma separated text or numbers.
    ''',
    'GEOM': '''Specify a geometry in GeoJSON format either as an inline value,
    stdin, or a file. ``@-`` specifies stdin and ``@filename`` specifies
    reading from a file named 'filename'. Otherwise, the value is assumed to
    be GeoJSON.
    ''',
    'FILTER': '''Specify a Data API search filter provided as JSON.
    ``@-`` specifies stdin and ``@filename`` specifies reading from a file
    named 'filename'. Otherwise, the value is assumed to be JSON.
    ''',
    'ITEM-TYPE': '''Specify Item-Type(s) of interest. Case-insensitive,
    supports glob-matching, e.g. ``psscene*`` means ``PSScene3Band`` and
    ``PSScene4Band``. The ``all`` value specifies every Item-Type.
    ''',
    'ASSET-TYPE': '''Specify Asset-Type(s) of interest. Case-insenstive,
    supports glob-matching, e.g. ``visual*`` specifies ``visual`` and
    ``visual_xml``.
    '''
}


class _LenientChoice(click.Choice):
    '''Like click.Choice but allows
    case-insensitive prefix matching
    optional 'all' matching
    optional prefix matching
    glob matching
    format fail msges for large selection of choices

    returns a list unlike choice (to support 'all')
    '''

    allow_all = False
    allow_prefix = False

    def get_metavar(self, param):
        return self.name.upper()

    def _fail(self, msg, val, param, ctx):
        self.fail('%s choice: %s.\nChoose from:\n\t%s' %
                  (msg, val, '\n\t'.join(self.choices)), param, ctx)

    def convert(self, val, param, ctx):
        matched = set()
        for p in _split(val):
            matched = matched.union(self._match(p, param, ctx))
        return list(matched)

    def _match(self, val, param, ctx, update_on_fail=True):
        lval = val.lower()
        if lval == 'all' and self.allow_all:
            return self.choices
        if '*' in lval:
            pat = lval.replace('*', '.*')
            matches = [c for c in self.choices
                       if re.match(pat, c.lower())]
        elif self.allow_prefix:
            matches = [c for c in self.choices
                       if c.lower().startswith(lval)]
        else:
            matches = [c for c in self.choices if c.lower() == lval]
        if not matches and update_on_fail:
            self._update_choices()  # get choices from a remote source.
            return self._match(val, param, ctx, False)  # Try one more time
        elif not matches:
            self._fail('invalid', val, param, ctx)
        else:
            return matches

    def _update_choices(self):
        self.choices = self.get_remote_choices()

    def get_remote_choices(self):
        '''
        No-op just returns current choices.
        Subclasses should override.
        '''
        return self.choices


class Bundle(_LenientChoice):
    name = 'bundle'
    allow_all = True

    def __init__(self):
        _LenientChoice.__init__(self, DEFAULT_BUNDLES)

    def get_remote_choices(self):
        return get_bundles()


class ItemType(_LenientChoice):
    name = 'item-type'
    allow_all = True
    allow_prefix = True

    def __init__(self):
        _LenientChoice.__init__(self, DEFAULT_ITEM_TYPES)

    def get_remote_choices(self):
        return get_item_types()


class AssetType(_LenientChoice):
    name = 'asset-type'

    def __init__(self):
        _LenientChoice.__init__(self, DEFAULT_ASSET_TYPES)

    def get_remote_choices(self):
        return get_asset_types()


class AssetTypePerm(AssetType):

    @staticmethod
    def to_permissions(asset_types):
        return filters.permission_filter(*[
            'assets.%s:download' % a for a in asset_types
        ])

    def convert(self, val, param, ctx):
        return AssetTypePerm.to_permissions(
            AssetType.convert(self, val, param, ctx))


class _FilterFieldValues(CompositeParamType):
    name = 'field values'
    arity = 2

    def convert(self, val, param, ctx):
        field, vals = val
        vals = _split(vals)
        parsed = []
        for v in vals:
            v = v.strip()
            if not v:
                continue
            try:
                parsed.append(self.val_type(v))
            except ValueError:
                self.fail('invalid value: %s' % v, param, ctx)
        return self._builder(field, *parsed)


class StringIn(_FilterFieldValues):
    val_type = str

    @property
    def _builder(self):
        return filters.string_filter


class NumberIn(_FilterFieldValues):
    val_type = float

    @property
    def _builder(self):
        return filters.num_filter


class Range(CompositeParamType):
    arity = 3
    name = 'field comp value'
    comp_ops = ['lt', 'lte', 'gt', 'gte']

    @property
    def _builder(self):
        return filters.range_filter

    def _parse(self, val, param, ctx):
        try:
            return float(val)
        except ValueError:
            self.fail('invalid value for range: "%s", must be number' % val)

    def convert(self, vals, param, ctx):
        field, comp_op, val = vals
        if comp_op not in self.comp_ops:
            self.fail('invalid operator: %s. allowed: %s' % (
                comp_op, ','.join(self.comp_ops)), param, ctx)
        args = dict([(comp_op, self._parse(val, param, ctx))])
        return self._builder(field, **args)


class DateRange(Range):
    @property
    def _builder(self):
        return filters.date_range

    def _parse(self, val, param, ctx):
        parsed = strp_lenient(val)
        if parsed is None:
            self.fail('invalid date: %s.' % val, param, ctx)
        return parsed


class GeomFilter(click.ParamType):
    name = 'geom'

    def convert(self, val, param, ctx):
        val = read(val)
        if not val:
            return []
        try:
            geoj = json.loads(val)
        except ValueError:
            raise click.BadParameter('invalid GeoJSON')
        geom = geometry_from_json(geoj)
        if geom is None:
            raise click.BadParameter('unable to find geometry in input')
        return [filters.geom_filter(geom)]


class FilterJSON(click.ParamType):
    name = 'filter'

    def convert(self, val, param, ctx):
        val = read(val)
        if not val:
            return []
        try:
            filt = json.loads(val)
        except ValueError:
            raise click.BadParameter('invalid JSON')
        if not filters.is_filter_like(filt):
            self.fail('Does not appear to be valid filter', param, ctx)
        return filt


class SortSpec(CompositeParamType):
    name = 'field order'
    arity = 2

    def convert(self, val, param, ctx):
        if not val:
            return ''
        field, order = val
        fields = ('published', 'acquired')
        if field not in fields:
            raise click.BadParameter(
                'sort only supports: %s' % ' '.join(fields))
        orders = ('asc', 'desc')
        if order not in orders:
            raise click.BadParameter(
                'order only supports: %s' % ' '.join(orders))
        return ' '.join(val)


class BoundingBox(click.ParamType):
    name = 'bbox'

    def convert(self, val, param, ctx):
        try:
            xmin, ymin, xmax, ymax = map(float, val.split(','))
        except (TypeError, ValueError):
            raise click.BadParameter('Invalid bounding box')
        return (xmin, ymin, xmax, ymax)


class DateInterval(click.ParamType):
    name = 'date interval'

    def convert(self, val, param, ctx):
        dates = val.split('/')
        if len(dates) > 2:
            raise click.BadParameter('Too many dates')

        for date in dates:
            if date != '..' and strp_lenient(date) is None:
                raise click.BadParameter('Invalid date: {}'.format(date))


class ClipAOI(click.ParamType):
    name = 'clip'

    def convert(self, val, param, ctx):
        val = read(val)
        if not val:
            return []
        try:
            json.loads(val)
        except ValueError:
            raise click.BadParameter('invalid GeoJSON')
        return val


class RequiredUnless(click.Option):
    def __init__(self, *args, **kwargs):
        self.this_opt_exists = kwargs.pop('this_opt_exists')
        super(RequiredUnless, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if self.name not in opts and self.this_opt_exists not in opts:
            raise click.UsageError('{} is required.'.format(self.name))
        return super(RequiredUnless, self).handle_parse_result(ctx, opts, args)



# Copyright 2015 Planet Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from datetime import datetime
from . import exceptions
import json
import mimetypes
import os
import random
import re
import string
import threading
from requests.compat import urlparse
from ._fatomic import atomic_open

_ISO_FMT = '%Y-%m-%dT%H:%M:%S.%f+00:00'


def _planet_json_file():
    return os.path.join(os.path.expanduser('~'), '.planet.json')


def read_planet_json():
    fname = _planet_json_file()
    contents = {}
    if os.path.exists(fname):
        with open(fname, 'r') as fp:
            contents = json.loads(fp.read())
    return contents


def write_planet_json(contents):
    fname = _planet_json_file()
    with atomic_open(fname, 'w') as fp:
        fp.write(json.dumps(contents))


def geometry_from_json(obj):
    '''try to find a geometry in the provided JSON object'''
    obj_type = obj.get('type', None)
    if not obj_type:
        return None
    if obj_type == 'FeatureCollection':
        features = obj.get('features', [])
        if len(features):
            obj = obj['features'][0]
            obj_type = obj.get('type', None)
        else:
            return None
    if obj_type == 'Feature':
        geom = obj['geometry']
    else:
        geom = obj
    # @todo we're just assuming it's a geometry at this point
    if 'coordinates' in geom:
        return geom


def check_status(response):
    '''check the status of the response and if needed raise an APIException'''
    status = response.status_code
    if status < 300:
        return
    exception = {
        400: exceptions.BadQuery,
        401: exceptions.InvalidAPIKey,
        403: exceptions.NoPermission,
        404: exceptions.MissingResource,
        429: exceptions.TooManyRequests,
        500: exceptions.ServerError
    }.get(status, None)

    # differentiate between over quota and rate-limiting
    if status == 429 and 'quota' in response.text.lower():
        exception = exceptions.OverQuota

    if exception:
        raise exception(response.text)

    raise exceptions.APIException('%s: %s' % (status, response.text))


def get_filename(response):
    """Derive a filename from the given response.

    >>> import requests
    >>> from planet.api import utils
    >>> response = requests.Response()
    >>> response.headers = {
    ...     'date': 'Thu, 14 Feb 2019 16:13:26 GMT',
    ...     'last-modified': 'Wed, 22 Nov 2017 17:22:31 GMT',
    ...     'accept-ranges': 'bytes',
    ...     'content-type': 'image/tiff',
    ...     'content-length': '57350256',
    ...     'content-disposition': 'attachment; filename="open_california.tif"'
    ... }
    >>> response.url = 'https://planet.com/path/to/example.tif?foo=f6f1'
    >>> print(utils.get_filename(response))
    open_california.tif
    >>> del response
    >>> response = requests.Response()
    >>> response.headers = {
    ...     'date': 'Thu, 14 Feb 2019 16:13:26 GMT',
    ...     'last-modified': 'Wed, 22 Nov 2017 17:22:31 GMT',
    ...     'accept-ranges': 'bytes',
    ...     'content-type': 'image/tiff',
    ...     'content-length': '57350256'
    ... }
    >>> response.url = 'https://planet.com/path/to/example.tif?foo=f6f1'
    >>> print(utils.get_filename(response))
    example.tif
    >>> del response
    >>> response = requests.Response()
    >>> response.headers = {
    ...     'date': 'Thu, 14 Feb 2019 16:13:26 GMT',
    ...     'last-modified': 'Wed, 22 Nov 2017 17:22:31 GMT',
    ...     'accept-ranges': 'bytes',
    ...     'content-type': 'image/tiff',
    ...     'content-length': '57350256'
    ... }
    >>> response.url = 'https://planet.com/path/to/oops/'
    >>> print(utils.get_filename(response)) #doctest:+SKIP
    planet-bFL6pwki.tif
    >>>

    :param response: An HTTP response.
    :type response: :py:class:`requests.Response`
    :returns: a filename (i.e. ``basename``)
    :rtype: str
    """
    name = (get_filename_from_headers(response.headers) or
            get_filename_from_url(response.url) or
            get_random_filename(response.headers.get('content-type')))
    return name


def get_filename_from_headers(headers):
    """Get a filename from the Content-Disposition header, if available.

    >>> from planet.api import utils
    >>> headers = {
    ...     'date': 'Thu, 14 Feb 2019 16:13:26 GMT',
    ...     'last-modified': 'Wed, 22 Nov 2017 17:22:31 GMT',
    ...     'accept-ranges': 'bytes',
    ...     'content-type': 'image/tiff',
    ...     'content-length': '57350256',
    ...     'content-disposition': 'attachment; filename="open_california.tif"'
    ... }
    >>> name = utils.get_filename_from_headers(headers)
    >>> print(name)
    open_california.tif
    >>>
    >>> headers.pop('content-disposition', None)
    'attachment; filename="open_california.tif"'
    >>> name = utils.get_filename_from_headers(headers)
    >>> print(name)
    None
    >>>

    :param headers dict: a ``dict`` of response headers
    :returns: a filename (i.e. ``basename``)
    :rtype: str or None
    """
    cd = headers.get('content-disposition', '')
    match = re.search('filename="?([^"]+)"?', cd)
    return match.group(1) if match else None


def get_filename_from_url(url):
    """Get a filename from a URL.

    >>> from planet.api import utils
    >>> urls = [
    ...     'https://planet.com/',
    ...     'https://planet.com/path/to/',
    ...     'https://planet.com/path/to/example.tif',
    ...     'https://planet.com/path/to/example.tif?foo=f6f1&bar=baz',
    ...     'https://planet.com/path/to/example.tif?foo=f6f1&bar=baz#quux'
    ... ]
    >>> for url in urls:
    ...     print('{} -> {}'.format(url, utils.get_filename_from_url(url)))
    ...
    https://planet.com/ -> None
    https://planet.com/path/to/ -> None
    https://planet.com/path/to/example.tif -> example.tif
    https://planet.com/path/to/example.tif?foo=f6f1&bar=baz -> example.tif
    https://planet.com/path/to/example.tif?foo=f6f1&bar=baz#quux -> example.tif
    >>>

    :returns: a filename (i.e. ``basename``)
    :rtype: str or None
    """
    path = urlparse(url).path
    name = path[path.rfind('/')+1:]
    return name or None


def get_random_filename(content_type=None):
    """Get a pseudo-random, Planet-looking filename.

    >>> from planet.api import utils
    >>> print(utils.get_random_filename()) #doctest:+SKIP
    planet-61FPnh7K
    >>> print(utils.get_random_filename('image/tiff')) #doctest:+SKIP
    planet-V8ELYxy5.tif
    >>>

    :returns: a filename (i.e. ``basename``)
    :rtype: str
    """
    extension = mimetypes.guess_extension(content_type or '') or ''
    characters = string.ascii_letters + '0123456789'
    letters = ''.join(random.sample(characters, 8))
    name = 'planet-{}{}'.format(letters, extension)
    return name


def write_to_file(directory=None, callback=None, overwrite=True):
    '''Create a callback handler for asynchronous Body handling.

    If provided, the callback will be invoked as described in
    :py:meth:`planet.api.models.Body.write`. In addition, if the download
    is skipped because the destination exists, the callback will be invoked
    with ``callback(skip=body)``.

    The name of the file written to will be determined from the Body.name
    property.

    :param directory str: The optional directory to write to.
    :param callback func: An optional callback to receive notification of
                          write progress.
    :param overwrite bool: Overwrite any existing files. Defaults to True.
    '''

    def writer(body):
        file = os.path.join(directory or '.', body.name)
        if overwrite or not os.path.exists(file):
            body.write(file, callback)
        else:
            if callback:
                callback(skip=body)
            body.response.close()
    return writer


def strp_timestamp(value):
    return datetime.strptime(value, _ISO_FMT)


def strf_timestamp(when):
    return datetime.strftime(when, _ISO_FMT)


def strp_lenient(when):
    when = when[:-1] if when[-1] == 'Z' else when
    for i in range(0, 9):
        try:
            return datetime.strptime(when, _ISO_FMT[:i*-3 or len(_ISO_FMT)])
        except ValueError:
            pass


class GeneratorAdapter(list):
    '''Allow a generator to be used in JSON serialization'''
    def __init__(self, gen):
        self.gen = gen

    def __iter__(self):
        return self.gen

    def __len__(self):
        return 1


def probably_wkt(text):
    '''Quick check to determine if the provided text looks like WKT'''
    valid = False
    valid_types = set([
        'POINT', 'LINESTRING', 'POLYGON', 'MULTIPOINT',
        'MULTILINESTRING', 'MULTIPOLYGON', 'GEOMETRYCOLLECTION',
    ])
    matched = re.match(r'(\w+)\s*\([^)]+\)', text.strip())
    if matched:
        valid = matched.group(1).upper() in valid_types
    return valid


def probably_geojson(input):
    '''A quick check to see if this input looks like GeoJSON. If not a dict
    JSON-like object, attempt to parse input as JSON. If the resulting object
    has a type property that looks like GeoJSON, return that object or None'''
    valid = False
    if not isinstance(input, dict):
        try:
            input = json.loads(input)
        except ValueError:
            return None
    typename = input.get('type', None)
    supported_types = set([
        'Point', 'LineString', 'Polygon', 'MultiPoint', 'MultiLineString',
        'MultiPolygon', 'GeometryCollection', 'Feature', 'FeatureCollection'
    ])
    valid = typename in supported_types
    return input if valid else None


def handle_interrupt(cancel, f, *a, **kw):
    '''Execute a function f(*a, **kw) listening for KeyboardInterrupt and if
    handled, invoke the cancel function. Blocks until f is complete or the
    interrupt is handled.
    '''
    res = []

    def run():
        res.append(f(*a, **kw))
    t = threading.Thread(target=run)
    t.start()
    # poll (or we miss the interrupt) and await completion
    try:
        while t.is_alive():
            t.join(.1)
    except KeyboardInterrupt:
        cancel()
        raise
    return res and res.pop()



# Copyright 2017 Planet Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from .types import (
    AssetType,
    AssetTypePerm,
    DateRange,
    GeomFilter,
    FilterJSON,
    ItemType,
    NumberIn,
    Range,
    SortSpec,
    StringIn,
    Bundle
)


pretty = click.option('-pp/-r', '--pretty/--no-pretty', default=None,
                      is_flag=True, help='Format JSON output')

dest_dir = click.option('-d', '--dest', help='Destination directory',
                        type=click.Path(file_okay=False, resolve_path=True,
                                        exists=True))

num_type = type('number', (click.types.IntParamType,), {'name': 'number'})()


def limit_option(default):
    limit_msg = "Limit the number of items. Default: {}".format(default)
    return click.option('--limit', default=default, required=False,
                        type=num_type, help=limit_msg)


geom_filter = click.option('--geom', type=GeomFilter(), help=(
    'Specify a geometry filter as geojson.'
))

date_range_filter = click.option(
    '--date', nargs=3, multiple=True, type=DateRange(),
    help=(
        'Filter field by date.'
    )
)

range_filter = click.option(
    '--range', multiple=True, type=Range(),
    help=(
        'Filter field by numeric range.'
    )
)

number_in_filter = click.option(
    '--number-in', multiple=True, type=NumberIn(),
    help=(
        'Filter field by numeric in.'
    )
)

string_in_filter = click.option(
    '--string-in', multiple=True, type=StringIn(),
    help=(
        'Filter field by string in.'
    )
)

item_type_option = click.option(
    '--item-type', multiple=True, required=True, type=ItemType(), help=(
       'Specify item type(s)'
    )
)

bundle_option = click.option(
    '--bundle', multiple=False, required=True, type=Bundle(), help=(
       'Specify bundle(s)'
    )
)

asset_type_option = click.option(
    '--asset-type', multiple=True, required=True, type=AssetType(), help=(
        'Specify asset type(s)'
    )
)

asset_type_perms = click.option(
    '--asset-type', multiple=True, required=False, type=AssetTypePerm(), help=(
        'Specify asset type(s) permissions'
    )
)

# @todo add validate/parse
filter_json_option = click.option(
    '--filter-json', default='@-', type=FilterJSON(), help=(
        'Use the specified filter'
    )
)

ndjson_option = click.option('--ndjson', is_flag=True, help=(
    'Request output as new-line delimited json.'
))


sort_order = click.option(
    '--sort', type=SortSpec(), help=(
        'Specify sort ordering as published/acquired asc/desc'
    )
)

_filter_opts = [date_range_filter, range_filter, number_in_filter,
                string_in_filter, geom_filter, filter_json_option]


def filter_opts(fun):
    '''Decorator for all search filter options'''
    for o in _filter_opts:
        fun = o(fun)
    return fun


def search_request_opts(fun):
    '''Decorator for common search request options'''
    return sort_order(item_type_option(filter_opts(fun)))



# Copyright 2017-2019 Planet Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from itertools import chain
import json
import logging
import re
from os import path
import sys
import tempfile
import textwrap
import threading
import time

import click
from click import termui

from planet import api
from planet.api import filters


def _split(value):
    '''return input split on any whitespace or comma'''
    return re.split(r'\s+|,', value)


def and_filter_from_opts(opts):
    '''build an AND filter from the provided opts dict as passed to a command
    from the filter_options decorator. Assumes all dict values are lists of
    filter dict constructs.'''
    return filters.and_filter(*list(chain.from_iterable([
        o for o in opts.values() if o]
    )))


def check_writable(dirpath):
    try:
        tempfile.NamedTemporaryFile(dir=dirpath).close()
    except OSError:
        return False
    # in windows with a vagrant ro-mount, this was raised instead
    except IOError:
        return False
    return True


def filter_from_opts(**kw):
    '''Build a AND filter from the provided kwargs defaulting to an
    empty 'and' filter (@todo: API workaround) if nothing is provided.

    If the 'filter_json' argument is provided, this will be assumed to contain
    a filter specification and will be anded with other filters. If the
    'filter_json' is a search, the search filter value will be used.

    All kw values should be tuple or list
    '''
    filter_in = kw.pop('filter_json', None)
    active = and_filter_from_opts(kw)
    if filter_in:
        filter_in = filter_in.get('filter', filter_in)
        if len(active['config']) > 0:
            active = filters.and_filter(active, filter_in)
        else:
            active = filter_in
    return active


def search_req_from_opts(**kw):
    # item_type will be list of lists - flatten
    item_types = chain.from_iterable(kw.pop('item_type'))
    name = kw.pop('name', '')
    interval = kw.pop('interval', '')
    filt = filter_from_opts(**kw)
    return filters.build_search_request(
        filt, item_types, name=name, interval=interval)


def create_order_request(**kwargs):
    for opt in ('item_type', 'bundle'):
        inputvalue = kwargs.get(opt)
        if len(inputvalue) > 1:
            raise click.ClickException(
                'only one value for {} is allowed.'.format(opt))

    item_type = kwargs.get('item_type')[0]
    bundle = kwargs.get('bundle')[0]
    ids = kwargs.get('id').split(',')
    email = kwargs.get('email')
    archive = kwargs.get('zip')
    config = kwargs.get('cloudconfig')
    clip = kwargs.get('clip')
    tools = kwargs.get('tools')

    request = {'name': kwargs.get('name'),
               'products': [{'item_ids': ids,
                             'item_type': item_type,
                             'product_bundle': bundle}
                            ],
               'tools': [
    ],
        'delivery': {
    },
        'notifications': {
                   'email': email
    },
    }

    if archive is not None:
        request["delivery"]["archive_filename"] = "{{name}}_{{order_id}}.zip"
        request["delivery"]["archive_type"] = "zip"

        # If single_archive is not set, each bundle will be zipped, as opposed
        # to the entire order.
        if archive == "order":
            request["delivery"]["single_archive"] = True

    if config:
        with open(config, 'r') as f:
            conf = json.load(f)
            request["delivery"].update(conf)

    # NOTE clip is the only tool that currently can be specified via CLI param.
    # A full tool chain can be specified via JSON file, so that will overwrite
    # clip if both are present. TODO add other common tools as params.
    if clip and not tools:
        toolchain = [{'clip': {'aoi': json.loads(clip)}}]
        request['tools'].extend(toolchain)

    if tools:
        with open(tools, 'r') as f:
            toolchain = json.load(f)
            request["tools"].extend(toolchain)

    return request


def call_and_wrap(func, *args, **kw):
    '''call the provided function and wrap any API exception with a click
    exception. this means no stack trace is visible to the user but instead
    a (hopefully) nice message is provided.
    note: could be a decorator but didn't play well with click
    '''
    try:
        return func(*args, **kw)
    except api.exceptions.APIException as ex:
        click_exception(ex)


def click_exception(ex):
    if type(ex) is api.exceptions.APIException:
        raise click.ClickException('Unexpected response: %s' % str(ex))
    msg = "%s: %s" % (type(ex).__name__, str(ex))
    raise click.ClickException(msg)


def echo_json_response(response, pretty, limit=None, ndjson=False):
    '''Wrapper to echo JSON with optional 'pretty' printing. If pretty is not
    provided explicity and stdout is a terminal (and not redirected or piped),
    the default will be to indent and sort keys'''
    indent = None
    sort_keys = False
    nl = False
    if not ndjson and (pretty or (pretty is None and sys.stdout.isatty())):
        indent = 2
        sort_keys = True
        nl = True
    try:
        if ndjson and hasattr(response, 'items_iter'):
            items = response.items_iter(limit)
            for item in items:
                click.echo(json.dumps(item))
        elif not ndjson and hasattr(response, 'json_encode'):
            response.json_encode(click.get_text_stream('stdout'), limit=limit,
                                 indent=indent, sort_keys=sort_keys)
        else:
            res = response.get_raw()
            if len(res) == 0:  # if the body is empty, just return the status
                click.echo("status: {}".format(response.response.status_code))
            else:
                res = json.dumps(json.loads(res), indent=indent,
                                 sort_keys=sort_keys)
                click.echo(res)
            if nl:
                click.echo()
    except IOError as ioe:
        # hide scary looking broken pipe stack traces
        raise click.ClickException(str(ioe))


def read(value, split=False):
    '''Get the value of an option interpreting as a file implicitly or
    explicitly and falling back to the value if not explicitly specified.
    If the value is '@name', then a file must exist with name and the returned
    value will be the contents of that file. If the value is '@-' or '-', then
    stdin will be read and returned as the value. Finally, if a file exists
    with the provided value, that file will be read. Otherwise, the value
    will be returned.
    '''
    v = str(value)
    retval = value
    if v[0] == '@' or v == '-':
        fname = '-' if v == '-' else v[1:]
        try:
            with click.open_file(fname) as fp:
                if not fp.isatty():
                    retval = fp.read()
                else:
                    retval = None
        # @todo better to leave as IOError and let caller handle it
        # to better report in context of call (e.g. the option/type)
        except IOError as ioe:
            # if explicit and problems, raise
            if v[0] == '@':
                raise click.ClickException(str(ioe))
    elif path.exists(v) and path.isfile(v):
        with click.open_file(v) as fp:
            retval = fp.read()
    if retval and split and type(retval) != tuple:
        retval = _split(retval.strip())
    return retval


class _BaseOutput(object):

    refresh_rate = 1

    def _report_complete(self, item, asset, path=None):
        msg = {
            'item': item['id'],
            'asset': asset['type'],
            'location': path or asset['location']
        }
        # cancel() allows report log to persist for both ANSI & regular output
        self.cancel()
        click.echo(json.dumps(msg))

    def __init__(self, thread, dl):
        self._thread = thread
        self._timer = None
        self._dl = dl
        self._running = False
        dl.on_complete = self._report_complete

    def _schedule(self):
        if self._thread.is_alive() and self._running:
            self._timer = threading.Timer(self.refresh_rate, self._run)
            self._timer.start()
            return True

    def _run(self, exit=False):
        if self._running:
            self._output(self._dl.stats())
        if not exit and self._running and not self._schedule():
            self._run(True)

    def start(self):
        self._running = True
        self._run()

    def cancel(self):
        self._running = False
        self._timer and self._timer.cancel()


class Output(_BaseOutput):

    def _output(self, stats):
        logging.info('%s', stats)


class AnsiOutput(_BaseOutput):

    def __init__(self, *args, **kw):
        _BaseOutput.__init__(self, *args, **kw)
        self._start = time.time()
        # log msg ring buffer
        self._records = deque(maxlen=100)
        self._lock = threading.Lock()
        self._stats = {}

        # highjack the root handler, remove existing and replace with one
        # that feeds our ring buffer
        h = logging.Handler()
        root = logging.getLogger('')
        h.formatter = root.handlers[0].formatter
        h.emit = self._emit
        root.handlers = (h,)
        self._handler = h

    def start(self):
        click.clear()
        _BaseOutput.start(self)

    def _emit(self, record):
        with self._lock:
            self._records.append(self._handler.format(record))
            self._do_output()

    def _output(self, stats):
        with self._lock:
            self._stats.update(stats)
            self._do_output()

    def _do_output(self):
        # renders a terminal like:
        # highlighted status rows
        # ....
        #
        # scrolling log output
        # ...
        width, height = click.termui.get_terminal_size()
        wrapper = textwrap.TextWrapper(width=width)
        self._stats['elapsed'] = '%d' % (time.time() - self._start)
        stats = ['%s: %s' % (k, v) for k, v in sorted(self._stats.items())]
        stats = wrapper.wrap(''.join([s.ljust(25) for s in stats]))
        remaining = height - len(stats) - 2
        stats = [s.ljust(width) for s in stats]
        lidx = max(0, len(self._records) - remaining)
        loglines = []
        while remaining > 0 and lidx < len(self._records):
            wrapped = wrapper.wrap(self._records[lidx])
            while remaining and wrapped:
                loglines.append(wrapped.pop(0))
                remaining -= 1
            lidx += 1
        # clear/cursor-to-1,1/hightlight
        click.echo(u'\u001b[2J\u001b[1;1H\u001b[30;47m' + '\n'.join(stats)
                   # unhighlight
                   + u'\u001b[39;49m\n' + '\n'.join(loglines))


def downloader_output(dl, disable_ansi=False):
    thread = threading.current_thread()
    # do fancy output if we can or not explicitly disabled
    if sys.stdout.isatty() and not disable_ansi and not termui.WIN:
        return AnsiOutput(thread, dl)
    # work around for lack of nice output for downloader on windows:
    # unless told to be quiet, set logging higher to get some output
    # @todo fallback to simpler 'UI' when isatty on win
    if termui.WIN and not disable_ansi:
        logging.getLogger('').setLevel(logging.INFO)
    return Output(thread, dl)


def ids_from_search_response(resp):
    ret = []
    r = json.loads(resp)
    for feature in r['features']:
        ret.append(feature['id'])
    return ','.join(ret)



# Copyright 2015 Planet Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
from .dispatch import RequestsDispatcher
from . import auth
from .exceptions import (InvalidIdentity, APIException)
from . import models
from . import filters


class _Base(object):
    '''High-level access to Planet's API.'''

    def __init__(self, api_key=None, base_url='https://api.planet.com/',
                 workers=4):
        '''
        :param str api_key: API key to use. Defaults to environment variable.
        :param str base_url: The base URL to use. Not required.
        :param int workers: The number of concurrent download workers
        '''
        api_key = api_key or auth.find_api_key()
        self.auth = api_key and auth.APIKey(api_key)
        self.base_url = base_url
        if not self.base_url.endswith('/'):
            self.base_url += '/'
        self.dispatcher = RequestsDispatcher(workers)

    def shutdown(self):
        self.dispatcher._asyncpool.executor.shutdown(wait=False)

    def _url(self, path):
        if path.startswith('http'):
            url = path
        else:
            url = self.base_url + path
        return url

    def _request(self, path, body_type=models.JSON, params=None, auth=None):
        return models.Request(self._url(path), auth or self.auth, params,
                              body_type)

    def _get(self, path, body_type=models.JSON, params=None, callback=None):
        # convert any JSON objects to text explicitly
        for k, v in (params or {}).items():
            if isinstance(v, dict):
                params[k] = json.dumps(v)

        request = self._request(path, body_type, params)
        response = self.dispatcher.response(request)
        if callback:
            response.get_body_async(callback)
        return response

    def _download_many(self, paths, params, callback):
        return [self._get(path, models.Image, params=params, callback=callback)
                for path in paths]

    def login(self, identity, credentials):
        '''Login using email identity and credentials. Returns a JSON
        object containing an `api_key` property with the user's API_KEY.
        :param str identity: email
        :param str credentials: password
        :returns: JSON object (Python dict)
        '''
        result = self.dispatcher.session.post(self._url('v0/auth/login'),
                                              json={
                                                  'email': identity,
                                                  'password': credentials
                                              })
        status = result.status_code
        if status == 400:
            raise APIException('invalid parameters, login process has changed')
        elif status == 401:
            # do our best to get something out to the user
            msg = result.text
            try:
                msg = json.loads(result.text)['message']
            finally:
                raise InvalidIdentity(msg)
        elif status != 200:
            raise APIException('%s: %s' % (status, result.text))
        jwt = result.text
        payload = jwt.split('.')[1]
        rem = len(payload) % 4
        if rem > 0:
            payload += '=' * (4 - rem)
        payload = base64.urlsafe_b64decode(payload.encode('utf-8'))
        return json.loads(payload.decode('utf-8'))


def _patch_stats_request(request):
    '''If the request has no filter config, add one that should do what is
    expected (include all items)
    see: PE-11813
    '''
    filt = request.get('filter', {})
    if not filt.get('config', None):
        request['filter'] = filters.date_range('acquired',
                                               gt='1970-01-01T00:00:00Z')
    return request


class ClientV1(_Base):
    '''ClientV1 provides basic low-level access to Planet's API. Only one
    ClientV1 should be in existence for an application. The Client is thread
    safe and takes care to avoid API throttling and also retry any throttled
    requests. Most functions take JSON-like dict representations of API
    request bodies. Return values are usually a subclass of
    :py:class:`planet.api.models.Body`. Any exceptional http responses are
    handled by translation to one of the :py:mod:`planet.api.exceptions`
    classes.
    '''

    def _params(self, kw):
        params = {}
        if 'page_size' in kw:
            params['_page_size'] = kw['page_size']
        if 'sort' in kw and kw['sort']:
            params['_sort'] = ''.join(kw['sort'])
        if 'strict' in kw:
            # This transforms a Python boolean into a JSON boolean
            params['strict'] = json.dumps(kw['strict'])
        return params

    def create_search(self, request):
        '''Create a new saved search from the specified request.
        The request must contain a ``name`` property.

        :param request: see :ref:`api-search-request`
        :returns: :py:class:`planet.api.models.JSON`
        :raises planet.api.exceptions.APIException: On API error.
        '''
        body = json.dumps(request)
        return self.dispatcher.response(models.Request(
            self._url('data/v1/searches/'), self.auth,
            body_type=models.JSON, data=body, method='POST')).get_body()

    def quick_search(self, request, **kw):
        '''Execute a quick search with the specified request.

        :param request: see :ref:`api-search-request`
        :param `**kw`: See Options below
        :returns: :py:class:`planet.api.models.Items`
        :raises planet.api.exceptions.APIException: On API error.

        :Options:

        * page_size (int): Size of response pages
        * sort (string): Sorting order in the form `field (asc|desc)`
        '''
        body = json.dumps(request)
        params = self._params(kw)
        return self.dispatcher.response(models.Request(
            self._url('data/v1/quick-search'), self.auth, params=params,
            body_type=models.Items, data=body, method='POST')).get_body()

    def saved_search(self, sid, **kw):
        '''Execute a saved search by search id.

        :param sid string: The id of the search
        :returns: :py:class:`planet.api.models.Items`
        :raises planet.api.exceptions.APIException: On API error.

        :Options:

        * page_size (int): Size of response pages
        * sort (string): Sorting order in the form `field (asc|desc)`

        '''
        path = 'data/v1/searches/%s/results' % sid
        params = self._params(kw)
        return self._get(self._url(path), body_type=models.Items,
                         params=params).get_body()

    def get_searches(self, quick=False, saved=True):
        '''Get searches listing.

        :param quick bool: Include quick searches (default False)
        :param quick saved: Include saved searches (default True)
        :returns: :py:class:`planet.api.models.Searches`
        :raises planet.api.exceptions.APIException: On API error.
        '''
        params = {}
        if saved and not quick:
            params['search_type'] = 'saved'
        elif quick:
            params['search_type'] = 'quick'
        return self._get(self._url('data/v1/searches/'),
                         body_type=models.Searches, params=params).get_body()

    def stats(self, request):
        '''Get stats for the provided request.

        :param request dict: A search request that also contains the 'interval'
                             property.
        :returns: :py:class:`planet.api.models.JSON`
        :raises planet.api.exceptions.APIException: On API error.
        '''
        # work-around for API bug
        request = _patch_stats_request(request)
        body = json.dumps(request)
        return self.dispatcher.response(models.Request(
            self._url('data/v1/stats'), self.auth,
            body_type=models.JSON, data=body, method='POST')).get_body()

    def get_assets(self, item):
        '''Get the assets for the provided item representations.

        Item representations are obtained from search requests.

        :param request dict: An item representation from the API.
        :returns: :py:class:`planet.api.models.JSON`
        :raises planet.api.exceptions.APIException: On API error.
        '''
        assets_url = item['_links']['assets']
        return self._get(assets_url).get_body()

    def activate(self, asset):
        '''Request activation of the specified asset representation.

        Asset representations are obtained from :py:meth:`get_assets`.

        :param request dict: An asset representation from the API.
        :returns: :py:class:`planet.api.models.Body` with no response content
        :raises planet.api.exceptions.APIException: On API error.
        '''
        activate_url = asset['_links']['activate']
        return self._get(activate_url, body_type=models.Body).get_body()

    def download(self, asset, callback=None):
        '''Download the specified asset. If provided, the callback will be
        invoked asynchronously. Otherwise it is up to the caller to handle the
        response Body.

        :param asset dict: An asset representation from the API
        :param callback: An optional function to aysnchronsously handle the
                         download. See :py:func:`planet.api.write_to_file`
        :returns: :py:Class:`planet.api.models.Response` containing a
                  :py:Class:`planet.api.models.Body` of the asset.
        :raises planet.api.exceptions.APIException: On API error.
        '''
        download_url = asset['location']
        return self._get(download_url, models.Body, callback=callback)

    def get_item(self, item_type, id):
        '''Get the an item response for the given item_type and id

        :param item_type str: A valid item-type
        :param id str: The id of the item
        :returns: :py:Class:`planet.api.models.JSON`
        :raises planet.api.exceptions.APIException: On API error.
        '''
        url = 'data/v1/item-types/%s/items/%s' % (item_type, id)
        return self._get(url).get_body()

    def get_assets_by_id(self, item_type, id):
        '''Get an item's asset response for the given item_type and id

        :param item_type str: A valid item-type
        :param id str: The id of the item
        :returns: :py:Class:`planet.api.models.JSON`
        :raises planet.api.exceptions.APIException: On API error.
        '''
        url = 'data/v1/item-types/%s/items/%s/assets' % (item_type, id)
        return self._get(url).get_body()

    def get_mosaic_series(self, series_id):
        '''Get information pertaining to a mosaics series
        :returns: :py:Class:`planet.api.models.JSON`
        '''
        url = self._url('basemaps/v1/series/{}'.format(series_id))
        return self._get(url, models.JSON).get_body()

    def get_mosaics_for_series(self, series_id):
        '''Get list of mosaics available for a series
        :returns: :py:Class:`planet.api.models.Mosaics`
        '''
        url = self._url('basemaps/v1/series/{}/mosaics'.format(series_id))
        return self._get(url, models.Mosaics).get_body()

    def get_mosaics(self, name_contains=None):
        '''Get information for all mosaics accessible by the current user.

        :returns: :py:Class:`planet.api.models.Mosaics`
        '''
        params = {}
        if name_contains:
            params['name__contains'] = name_contains
        url = self._url('basemaps/v1/mosaics')
        return self._get(url, models.Mosaics, params=params).get_body()

    def get_mosaic_by_name(self, name):
        '''Get the API representation of a mosaic by name.

        :param name str: The name of the mosaic
        :returns: :py:Class:`planet.api.models.Mosaics`
        :raises planet.api.exceptions.APIException: On API error.
        '''
        params = {'name__is': name}
        url = self._url('basemaps/v1/mosaics')
        return self._get(url, models.Mosaics, params=params).get_body()

    def get_quads(self, mosaic, bbox=None):
        '''Search for quads from a mosaic that are inside the specified
        bounding box.  Will yield all quads if no bounding box is specified.

        :param mosaic dict: A mosaic representation from the API
        :param bbox tuple: A lon_min, lat_min, lon_max, lat_max area to search
        :returns: :py:Class:`planet.api.models.MosaicQuads`
        :raises planet.api.exceptions.APIException: On API error.
        '''
        if bbox is None:
            # Some bboxes can slightly exceed backend min/max latitude bounds
            xmin, ymin, xmax, ymax = mosaic['bbox']
            bbox = (max(-180, xmin), max(-85, ymin),
                    min(180, xmax), min(85, ymax))
        url = mosaic['_links']['quads']
        url = url.format(lx=bbox[0], ly=bbox[1], ux=bbox[2], uy=bbox[3])
        return self._get(url, models.MosaicQuads).get_body()

    def get_quad_by_id(self, mosaic, quad_id):
        '''Get a quad response for a specific mosaic and quad.

        :param mosaic dict: A mosaic representation from the API
        :param quad_id str: A quad id (typically <xcoord>-<ycoord>)
        :returns: :py:Class:`planet.api.models.JSON`
        :raises planet.api.exceptions.APIException: On API error.
        '''
        path = 'basemaps/v1/mosaics/{}/quads/{}'.format(mosaic['id'], quad_id)
        return self._get(self._url(path)).get_body()

    def get_quad_contributions(self, quad):
        '''Get information about which scenes contributed to a quad.

        :param quad dict: A quad representation from the API
        :returns: :py:Class:`planet.api.models.JSON`
        :raises planet.api.exceptions.APIException: On API error.
        '''
        url = quad['_links']['items']
        return self._get(url).get_body()

    def download_quad(self, quad, callback=None):
        '''Download the specified mosaic quad. If provided, the callback will
        be invoked asynchronously.  Otherwise it is up to the caller to handle
        the response Body.

        :param asset dict: A mosaic quad representation from the API
        :param callback: An optional function to aysnchronsously handle the
                         download. See :py:func:`planet.api.write_to_file`
        :returns: :py:Class:`planet.api.models.Response` containing a
                  :py:Class:`planet.api.models.Body` of the asset.
        :raises planet.api.exceptions.APIException: On API error.
        '''
        download_url = quad['_links']['download']
        return self._get(download_url, models.Body, callback=callback)

    def check_analytics_connection(self):
        '''
        Validate that we can use the Analytics API. Useful to test connectivity
        to test environments.
        :returns: :py:Class:`planet.api.models.JSON`
        '''
        return self._get(self._url('analytics/health')).get_body()

    def wfs_conformance(self):
        '''
        Details about WFS3 conformance
        :returns: :py:Class:`planet.api.models.JSON`
        '''
        return self._get(self._url('analytics/conformance')).get_body()

    def list_analytic_subscriptions(self, feed_id):
        '''
        Get subscriptions that the authenticated user has access to
        :param feed_id str: Return subscriptions associated with a particular
        feed only.
        :raises planet.api.exceptions.APIException: On API error.
        :returns: :py:Class:`planet.api.models.Subscriptions`
        '''
        params = {'feedID': feed_id}
        url = self._url('analytics/subscriptions')
        return self._get(url, models.Subscriptions, params=params).get_body()

    def get_subscription_info(self, subscription_id):
        '''
        Get the information describing a specific subscription.
        :param subscription_id:
        :raises planet.api.exceptions.APIException: On API error.
        :returns: :py:Class:`planet.api.models.JSON`
        '''
        url = self._url('analytics/subscriptions/{}'.format(subscription_id))
        return self._get(url, models.JSON).get_body()

    def list_analytic_feeds(self, stats):
        '''
        Get collections that the authenticated user has access to
        :raises planet.api.exceptions.APIException: On API error.
        :returns: :py:Class:`planet.api.models.Feeds`
        '''
        params = {'stats': stats}
        url = self._url('analytics/feeds')
        return self._get(url, models.Feeds, params=params).get_body()

    def get_feed_info(self, feed_id):
        '''
        Get the information describing a specific collection.
        :param subscription_id:
        :raises planet.api.exceptions.APIException: On API error.
        :returns: :py:Class:`planet.api.models.JSON`
        '''
        url = self._url('analytics/feeds/{}'.format(feed_id))
        return self._get(url, models.JSON).get_body()

    def list_analytic_collections(self):
        '''
        Get collections that the authenticated user has access to
        :raises planet.api.exceptions.APIException: On API error.
        :returns: :py:Class:`planet.api.models.WFS3Collections`
        '''
        params = {}
        url = self._url('analytics/collections')
        return self._get(url, models.WFS3Collections,
                         params=params).get_body()

    def get_collection_info(self, subscription_id):
        '''
        Get the information describing a specific collection.
        :param subscription_id:
        :raises planet.api.exceptions.APIException: On API error.
        :returns: :py:Class:`planet.api.models.JSON`
        '''
        url = 'analytics/collections/{}'.format(subscription_id)
        return self._get(self._url(url), models.JSON).get_body()

    def list_collection_features(self,
                                 subscription_id,
                                 bbox=None,
                                 time_range=None,
                                 before=None,
                                 after=None,
                                 ):
        '''
        List features for an analytic subscription.
        :param subscription_id:
        :param time_range str: ISO format datetime interval.
        :param bbox tuple: A lon_min, lat_min, lon_max, lat_max area to search
        :param before str: return features published before item with given ID
        :param after str: return features published after item with given ID
        :raises planet.api.exceptions.APIException: On API error.
        :returns: :py:Class:`planet.api.models.WFS3Features`
        '''
        params = {}

        if bbox:
            params['bbox'] = ','.join([str(b) for b in bbox])
        if time_range:
            params['time'] = time_range
        if before:
            params['before'] = before
        if after:
            params['after'] = after

        url = self._url('analytics/collections/{}/items'.format(
            subscription_id))
        return self._get(url, models.WFS3Features, params=params).get_body()

    def get_associated_resource_for_analytic_feature(self,
                                                     subscription_id,
                                                     feature_id,
                                                     resource_type):
        '''
        Get resource associated with some feature in an analytic subscription.
        Response might be JSON or a TIF, depending on requested resource.
        :param subscription_id str: ID of subscription
        :param feature_id str: ID of feature
        :param resource_type str: Type of resource to request.
        :raises planet.api.exceptions.APIException: On API error or resource
        type unavailable.
        :returns: :py:Class:`planet.api.models.JSON` for resource type
        `source-image-info`,  but can also return
        :py:Class:`planet.api.models.Response` containing a
        :py:Class:`planet.api.models.Body` of the resource.
        '''
        url = self._url(
            'analytics/collections/{}/items/{}/resources/{}'.format(
                subscription_id, feature_id, resource_type))
        response = self._get(url).get_body()
        return response

    def get_orders(self):
        '''Get information for all pending and completed order requests for
        the current user.

        :returns: :py:Class:`planet.api.models.Orders`
        '''

        # TODO filter 'completed orders', 'in progress orders', 'all orders'?
        url = self._url('compute/ops/orders/v2')
        orders = (self._get(url, models.Orders).get_body())
        return orders

    def get_individual_order(self, order_id):
        '''Get order request details by Order ID.

        :param order_id str: The ID of the Order
        :returns: :py:Class:`planet.api.models.Order`
        :raises planet.api.exceptions.APIException: On API error.
        '''
        url = self._url('compute/ops/orders/v2/{}'.format(order_id))
        return self._get(url, models.Order).get_body()

    def cancel_order(self, order_id):
        '''Cancel a running order by Order ID.

        :param order_id str: The ID of the Order to cancel
        :returns: :py:Class:`planet.api.models.Order`
        :raises planet.api.exceptions.APIException: On API error.
        '''
        url = self._url('compute/ops/orders/v2/{}'.format(order_id))
        return self.dispatcher.response(models.Request(url, self.auth,
                                                       body_type=models.Order,
                                                       method='PUT')
                                        ).get_body()

    def create_order(self, request):
        '''Create an order.

        :param asset:
        :returns: :py:Class:`planet.api.models.Response` containing a
                  :py:Class:`planet.api.models.Body` of the asset.
        :raises planet.api.exceptions.APIException: On API error.
        '''
        url = self._url('compute/ops/orders/v2')
        body = json.dumps(request)
        return self.dispatcher.response(models.Request(url, self.auth,
                                                       body_type=models.Order,
                                                       data=body,
                                                       method='POST')
                                        ).get_body()

    def download_order(self, order_id, callback=None):
        '''Download all items in an order.

        :param order_id: ID of order to download
        :returns: :py:Class:`planet.api.models.Response` containing a
                  :py:Class:`planet.api.models.Body` of the asset.
        :raises planet.api.exceptions.APIException: On API error.
        '''

        url = self._url('compute/ops/orders/v2/{}'.format(order_id))

        order = self._get(url, models.Order).get_body()
        locations = order.get_locations()

        return [self.download_location(location, callback=callback)
                for location in locations]

    def download_location(self, location, callback=None):
        '''Download an item in an order.

        :param location: location URL of item
        :returns: :py:Class:`planet.api.models.Response` containing a
                  :py:Class:`planet.api.models.Body` of the asset.
        :raises planet.api.exceptions.APIException: On API error.
        '''

        return self._get(location, models.JSON, callback=callback)



# Copyright 2017 Planet Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
from click.testing import CliRunner

from itertools import chain
import json
from .cli import (
    cli,
    clientv1,
)
from .opts import (
    asset_type_option,
    bundle_option,
    asset_type_perms,
    filter_opts,
    limit_option,
    pretty,
    search_request_opts,
    sort_order
)
from .types import (
    AssetTypePerm,
    BoundingBox,
    metavar_docs,
    DateInterval,
    ItemType,
    RequiredUnless,
    ClipAOI,
)
from .util import (
    call_and_wrap,
    check_writable,
    click_exception,
    filter_from_opts,
    downloader_output,
    echo_json_response,
    read,
    search_req_from_opts,
    create_order_request,
    ids_from_search_response,
)
from planet.api.utils import (
    handle_interrupt
)
from planet.api import downloader
from planet.api.utils import write_to_file

filter_opts_epilog = '\nFilter Formats:\n\n' + \
                     '\n'.join(['%s\n\n%s' % (k, v.replace('    ', '')
                                              .replace('``', '\''))
                                for k, v in metavar_docs.items()])

DEFAULT_SEARCH_LIMIT = 100
MAX_PAGE_SIZE = 250


@cli.group('data')
def data():
    '''Commands for interacting with the Data API'''
    pass


@data.command('filter', epilog=filter_opts_epilog)
@filter_opts
def filter_dump(**kw):
    '''Output a AND filter as JSON to stdout.

    If provided using --filter-json, combine the filters.

    The output is suitable for use in other commands via the
    --filter-json option.
    '''
    click.echo(json.dumps(filter_from_opts(**kw), indent=2))


@data.command('search', epilog=filter_opts_epilog)
@limit_option(DEFAULT_SEARCH_LIMIT)
@pretty
@asset_type_perms
@search_request_opts
def quick_search(limit, pretty, sort, **kw):
    '''Execute a quick search.'''
    req = search_req_from_opts(**kw)
    cl = clientv1()
    page_size = min(limit, MAX_PAGE_SIZE)
    echo_json_response(call_and_wrap(
        cl.quick_search, req, page_size=page_size, sort=sort
    ), pretty, limit)


@data.command('create-search', epilog=filter_opts_epilog)
@pretty
@click.option('--name', required=True)
@asset_type_perms
@search_request_opts
def create_search(pretty, **kw):
    '''Create a saved search'''
    req = search_req_from_opts(**kw)
    cl = clientv1()
    echo_json_response(call_and_wrap(cl.create_search, req), pretty)


@data.command('saved-search')
@click.argument('search_id', default='@-', required=False)
@sort_order
@pretty
@limit_option(DEFAULT_SEARCH_LIMIT)
def saved_search(search_id, sort, pretty, limit):
    '''Execute a saved search'''
    sid = read(search_id)
    cl = clientv1()
    page_size = min(limit, MAX_PAGE_SIZE)
    echo_json_response(call_and_wrap(
        cl.saved_search, sid, page_size=page_size, sort=sort
    ), limit=limit, pretty=pretty)


@data.command('searches')
@click.option('--quick', is_flag=True, help='Quick searches')
@click.option('--saved', default=True, is_flag=True,
              help='Saved searches (default)')
@limit_option(10)
def get_searches(quick, saved, limit):
    '''List searches'''
    cl = clientv1()
    response = call_and_wrap(cl.get_searches, quick, saved)
    echo_json_response(response, True, limit=limit)


@pretty
@asset_type_perms
@search_request_opts
@click.option('--interval', default='month',
              type=click.Choice(['hour', 'day', 'month', 'week', 'year']),
              help='Specify the interval to aggregate by.')
@data.command('stats', epilog=filter_opts_epilog)
def stats(pretty, **kw):
    '''Get search stats'''
    req = search_req_from_opts(**kw)
    cl = clientv1()
    echo_json_response(call_and_wrap(cl.stats, req), pretty)


def _disable_item_type(ctx, param, value):
    if value:
        for p in ctx.command.params:
            if p.name == 'item_type':
                p.required = False
    return value


@asset_type_option
@search_request_opts
@click.option('--search-id', is_eager=True, callback=_disable_item_type,
              type=str, help='Use the specified search')
@click.option('--dry-run', is_flag=True, help=(
    'Only report the number of items that would be downloaded.'
))
@click.option('--activate-only', is_flag=True, help=(
    'Only activate the items. Outputs URLS for downloading.'
))
@click.option('--quiet', is_flag=True, help=(
    'Disable ANSI control output'
))
@click.option('--dest', default='.', help=(
    'Location to download files to'), type=click.Path(
    exists=True, resolve_path=True, writable=True, file_okay=False))
@limit_option(None)
@data.command('download', epilog=filter_opts_epilog)
def download(asset_type, dest, limit, sort, search_id, dry_run, activate_only,
             quiet, **kw):
    '''Activate and download'''
    cl = clientv1()
    page_size = min(limit or MAX_PAGE_SIZE, MAX_PAGE_SIZE)
    asset_type = list(chain.from_iterable(asset_type))
    # even though we're using functionality from click.Path, this was needed
    # to detect inability to write on Windows in a read-only vagrant mount...
    # @todo check/report upstream
    if not activate_only and not check_writable(dest):
        raise click.ClickException(
            'download destination "%s" is not writable' % dest)
    if search_id:
        if dry_run:
            raise click.ClickException(
                'dry-run not supported with saved search')
        if any(kw[s] for s in kw):
            raise click.ClickException(
                'search options not supported with saved search')
        search, search_arg = cl.saved_search, search_id
    else:
        # any requested asset-types should be used as permission filters
        kw['asset_type'] = [AssetTypePerm.to_permissions(asset_type)]
        req = search_req_from_opts(**kw)
        if dry_run:
            req['interval'] = 'year'
            stats = cl.stats(req).get()
            item_cnt = sum([b['count'] for b in stats['buckets']])
            asset_cnt = item_cnt * len(asset_type)
            click.echo(
                'would download approximately %d assets from %s items' %
                (asset_cnt, item_cnt)
            )
            return
        else:
            search, search_arg = cl.quick_search, req

    dl = downloader.create(cl)
    output = downloader_output(dl, disable_ansi=quiet)
    # delay initial item search until downloader output initialized
    output.start()
    try:
        items = search(search_arg, page_size=page_size, sort=sort)
    except Exception as ex:
        output.cancel()
        click_exception(ex)
    func = dl.activate if activate_only else dl.download
    args = [items.items_iter(limit), asset_type]
    if not activate_only:
        args.append(dest)
    # invoke the function within an interrupt handler that will shut everything
    # down properly
    handle_interrupt(dl.shutdown, func, *args)


@cli.group('mosaics')
def mosaics():
    '''Commands for interacting with the Mosaics API'''
    pass


@mosaics.group('series')
def series():
    '''Commands for interacting with Mosaic Series through the Mosaics API'''
    pass


@series.command('describe')
@click.argument('series_id')
@pretty
def describe(series_id, pretty):
    cl = clientv1()
    echo_json_response(call_and_wrap(cl.get_mosaic_series, series_id), pretty)


@series.command('list-mosaics')
@click.argument('series_id')
@pretty
def list_mosaics_for_series(series_id, pretty):
    client = clientv1()
    series = client.get_mosaics_for_series(series_id)
    echo_json_response(series, pretty)


@mosaics.command('list')
@click.option('--prefix', default=None)
@pretty
def list_mosaics(pretty, prefix):
    '''List information for all available mosaics'''
    cl = clientv1()
    echo_json_response(call_and_wrap(cl.get_mosaics, prefix), pretty)


@mosaics.command('search')
@click.argument('name')
@click.option('--bbox', type=BoundingBox(), help=(
    'Region to query as a comma-delimited string:'
    ' lon_min,lat_min,lon_max,lat_max'
))
@click.option('--rbox', type=BoundingBox(), help='Alias for --bbox')
@limit_option(None)
@pretty
def search_mosaics(name, bbox, rbox, limit, pretty):
    '''Get quad IDs and information for a mosaic'''
    bbox = bbox or rbox
    cl = clientv1()
    mosaic, = cl.get_mosaic_by_name(name).items_iter(1)
    response = call_and_wrap(cl.get_quads, mosaic, bbox)
    echo_json_response(response, pretty, limit)


@mosaics.command('info')
@click.argument('name')
@pretty
def mosaic_info(name, pretty):
    '''Get information for a specific mosaic'''
    cl = clientv1()
    echo_json_response(call_and_wrap(cl.get_mosaic_by_name, name), pretty)


@mosaics.command('quad-info')
@click.argument('name')
@click.argument('quad')
@pretty
def quad_info(name, quad, pretty):
    '''Get information for a specific mosaic quad'''
    cl = clientv1()
    mosaic, = cl.get_mosaic_by_name(name).items_iter(1)
    echo_json_response(call_and_wrap(cl.get_quad_by_id, mosaic, quad), pretty)


@mosaics.command('contribution')
@click.argument('name')
@click.argument('quad')
@pretty
def quad_contributions(name, quad, pretty):
    '''Get contributing scenes for a mosaic quad'''
    cl = clientv1()
    mosaic, = cl.get_mosaic_by_name(name).items_iter(1)
    quad = cl.get_quad_by_id(mosaic, quad).get()
    response = call_and_wrap(cl.get_quad_contributions, quad)
    echo_json_response(response, pretty)


@mosaics.command('download')
@click.argument('name')
@click.option('--bbox', type=BoundingBox(), help=(
    'Region to download as a comma-delimited string:'
    ' lon_min,lat_min,lon_max,lat_max'
))
@click.option('--rbox', type=BoundingBox(), help='Alias for --bbox')
@click.option('--quiet', is_flag=True, help=(
    'Disable ANSI control output'
))
@click.option('--dest', default='.', help=(
    'Location to download files to'), type=click.Path(
    exists=True, resolve_path=True, writable=True, file_okay=False
))
@limit_option(None)
def download_quads(name, bbox, rbox, quiet, dest, limit):
    '''Download quads from a mosaic'''
    bbox = bbox or rbox
    cl = clientv1()

    dl = downloader.create(cl, mosaic=True)
    output = downloader_output(dl, disable_ansi=quiet)
    output.start()
    try:
        mosaic, = cl.get_mosaic_by_name(name).items_iter(1)
        items = cl.get_quads(mosaic, bbox).items_iter(limit)
    except Exception as ex:
        output.cancel()
        click_exception(ex)
    # invoke the function within an interrupt handler that will shut everything
    # down properly
    handle_interrupt(dl.shutdown, dl.download, items, [], dest)


@cli.group('analytics')
def analytics():
    '''Commands for interacting with the Analytics API'''
    pass


@analytics.command('check-connection')
@pretty
def health(pretty):
    '''Check that we can connect to the API'''
    cl = clientv1()
    click.echo('Using base URL: {}'.format(cl.base_url))
    response = cl.check_analytics_connection()
    echo_json_response(response, pretty)


@analytics.command('wfs-conformance')
@pretty
def conformance(pretty):
    '''
    Details about WFS3 conformance.
    :param pretty:
    :return:
    '''
    cl = clientv1()
    response = cl.wfs_conformance()
    echo_json_response(response, pretty)


@analytics.group('feeds')
def feeds():
    '''Commands for interacting with the Analytics API for collections'''
    pass


@feeds.command('list')
@limit_option(None)
@click.option('--stats', is_flag=True, default=False,
              help='Include feed stats')
@pretty
def list_feeds(pretty, limit, stats):
    '''List all subscriptions user has access to.'''
    cl = clientv1()
    response = cl.list_analytic_feeds(stats)
    echo_json_response(response, pretty, limit)


@feeds.command('list-mosaics')
@click.argument('feed_id')
def get_mosaic_list_for_feed(feed_id):
    '''List mosaics linked to feed'''
    cl = clientv1()
    feed_info = cl.get_feed_info(feed_id).get()

    for type_ in ['target', 'source']:
        feed_image_conf = feed_info.get(type_)

        if feed_image_conf['type'] != 'mosaic':
            msg_format = 'The {} for this feed is not a mosaic type.'
            click.ClickException(msg_format.format(type_))
            continue

        mosaic_series = feed_image_conf['config']['series_id']

        mosaics = cl.get_mosaics_for_series(mosaic_series)

        click.echo('{} mosaics:'.format(type_))
        for mosaic in mosaics.get()['mosaics']:
            click.echo('\t{}'.format(mosaic['name']))


@feeds.command('describe')
@click.argument('feed_id')
@pretty
def get_feed_info(feed_id, pretty):
    '''Get metadata for specific feed.'''
    cl = clientv1()
    feed_info = cl.get_feed_info(feed_id)
    echo_json_response(feed_info, pretty)


@analytics.group('subscriptions')
def subscriptions():
    '''
    Commands for interacting with the Analytics API for subscriptions
    '''
    pass


@subscriptions.command('list')
@click.option('--feed-id', type=str)
@limit_option(None)
@pretty
def list_subscriptions(pretty, limit, feed_id):
    '''List all subscriptions user has access to.'''
    cl = clientv1()
    response = cl.list_analytic_subscriptions(feed_id)
    echo_json_response(response, pretty, limit)


@subscriptions.command('list-mosaics')
@click.argument('subscription_id')
def get_mosaic_list_for_subscription(subscription_id):
    '''List mosaics linked to feed'''
    cl = clientv1()
    sub_info = cl.get_subscription_info(subscription_id).get()
    feed_info = cl.get_feed_info(sub_info['feedID']).get()

    for type_ in ['target', 'source']:
        feed_image_conf = feed_info.get(type_)

        if feed_image_conf['type'] != 'mosaic':
            msg_format = 'The {} for this feed is not a mosaic type.'
            click.ClickException(msg_format.format(type_))
            continue

        mosaic_series = feed_image_conf['config']['series_id']

        mosaics = cl.get_mosaics_for_series(mosaic_series)

        click.echo('{} mosaics:'.format(type_))
        for mosaic in mosaics.get()['mosaics']:
            click.echo('\t{}'.format(mosaic['name']))


@subscriptions.command('describe')
@click.argument('subscription_id')
@pretty
def get_subscription_info(subscription_id, pretty):
    '''Get metadata for specific subscription.'''
    cl = clientv1()
    sub_info = cl.get_subscription_info(subscription_id)
    echo_json_response(sub_info, pretty)


@analytics.group('collections')
def collections():
    '''Commands for interacting with the Analytics API for collections'''
    pass


@collections.command('list')
@limit_option(None)
@pretty
def list_collections(pretty, limit):
    '''List all collections user has access to.'''
    cl = clientv1()
    response = cl.list_analytic_collections()
    echo_json_response(response, pretty, limit)


@collections.command('list-mosaics')
@click.argument('subscription_id')
def get_mosaic_list_for_collection(subscription_id):
    '''List mosaics linked to feed'''
    cl = clientv1()
    sub_info = cl.get_subscription_info(subscription_id).get()
    feed_info = cl.get_feed_info(sub_info['feedID']).get()

    for type_ in ['target', 'source']:
        feed_image_conf = feed_info.get(type_)

        if feed_image_conf['type'] != 'mosaic':
            msg_format = 'The {} for this feed is not a mosaic type.'
            click.ClickException(msg_format.format(type_))
            continue

        mosaic_series = feed_image_conf['config']['series_id']

        mosaics = cl.get_mosaics_for_series(mosaic_series)

        click.echo('{} mosaics:'.format(type_))
        for mosaic in mosaics.get()['mosaics']:
            click.echo('\t{}'.format(mosaic['name']))


@collections.command('describe')
@click.argument('subscription_id')
@pretty
def get_collection_info(subscription_id, pretty):
    '''Get metadata for specific collection.'''
    cl = clientv1()
    sub_info = cl.get_collection_info(subscription_id)
    echo_json_response(sub_info, pretty)


@collections.command('resource-types')
@click.argument('subscription_id')
@pretty
def get_resource_types(subscription_id, pretty):
    '''Get available resource types.'''
    cl = clientv1()
    # Assumes that all features in a collection have the same list of
    # associated resource types
    features = cl.list_collection_features(subscription_id,
                                           None,
                                           None)
    feature_list = features.get()['features']
    if not feature_list:
        click.ClickException(
            'No features found, cannot determine resource types.').show()
        click.Abort()
    types = {item['rel'] for item in features.get()['features'][0]['links']}

    # The client and API only support these three, but there may be more link
    # types, ex. to things like tiles
    supported_types = {'source-image-info', 'target-quad', 'source-quad'}

    found_types = types.intersection(supported_types)
    click.echo('Found resource types: {}'.format(list(found_types)))


@collections.group('features')
def features():
    '''Commands for interacting with the Analytics API for features'''
    pass


@features.command('list')
@click.argument('subscription_id')
@click.option('--bbox', type=BoundingBox(), help=(
    'Region to query as a comma-delimited string:'
    ' lon_min,lat_min,lon_max,lat_max'
))
@click.option('--rbox', type=BoundingBox(), help='Alias for --bbox')
@click.option('--time-range', type=DateInterval(), help=(
    'Time interval. Can be open or closed interval, start times are '
    'inclusive and end times are exclusive: '
    '2019-01-01T00:00:00.00Z/2019-02-01T00:00:00.00Z (Closed interval for '
    'January 2019), 2019-01-01T00:00:00.00Z/.. (Open interval for all '
    'items since the start of January 2019), 2019-01-01T00:00:00.00Z '
    '(instant)'
))
@click.option('--before', type=str, help=(
    'Get results published before the item with the provided ID.'
))
@click.option('--after', type=str, help=(
    'Get results published after the item with the provided ID.'
))
@limit_option(100)
@pretty
def list_features(subscription_id, pretty, limit, rbox, bbox, time_range,
                  before, after):
    '''Request feature list for a particular subscription, 100 at a time.'''
    cl = clientv1()
    bbox = bbox or rbox
    features = cl.list_collection_features(subscription_id, bbox, time_range,
                                           before, after)
    echo_json_response(features, pretty, limit)


@features.command('list-all')
@click.argument('subscription_id')
@click.option('--bbox', type=BoundingBox(), help=(
    'Region to query as a comma-delimited string:'
    ' lon_min,lat_min,lon_max,lat_max'
))
@click.option('--rbox', type=BoundingBox(), help='Alias for --bbox')
@click.option('--time-range', type=DateInterval(), help=(
    'Time interval. Can be open or closed interval, start times are '
    'inclusive and end times are exclusive: '
    '2019-01-01T00:00:00.00Z/2019-02-01T00:00:00.00Z (Closed interval for '
    'January 2019), 2019-01-01T00:00:00.00Z/.. (Open interval for all '
    'items since the start of January 2019), 2019-01-01T00:00:00.00Z '
    '(instant)'
))
@click.option('--before', type=str, help=(
    'Get results published before the item with the provided ID.'
))
@click.option('--after', type=str, help=(
    'Get results published after the item with the provided ID.'
))
@pretty
def list_features_all(subscription_id, pretty, rbox, bbox, time_range, before,
                      after):
    '''Return every available feature for a particular subscription'''
    cl = clientv1()
    bbox = bbox or rbox
    features = cl.list_collection_features(subscription_id, bbox, time_range,
                                           before, after)
    echo_json_response(features, pretty)


@features.command('get')
@click.argument('resource_type', type=click.Choice(
    ['source-image-info', 'target-quad', 'source-quad']))
@click.argument('subscription_id')
@click.argument('feature_id')
@click.option('--dest', default='.', help=(
    'Location to download files to'), type=click.Path(
    exists=True, resolve_path=True, writable=True, file_okay=False
))
@pretty
def get_associated_resource(subscription_id, feature_id, resource_type, pretty,
                            dest):
    '''Request resources for a particular subscription/feature combination.'''
    cl = clientv1()
    if resource_type in ['target-quad', 'source-quad']:
        msg_format = 'Requesting {} for {}/{}, destination directory is: {}'
        click.echo(msg_format.format(
            resource_type,
            subscription_id,
            feature_id,
            dest
        ))

    resource = cl.get_associated_resource_for_analytic_feature(subscription_id,
                                                               feature_id,
                                                               resource_type)

    if resource_type == 'source-image-info':
        echo_json_response(resource, pretty)

    if resource_type in ['target-quad', 'source-quad']:
        writer = write_to_file(dest, None)
        writer(resource)
        click.echo('{} written, available at: {}/{}'.format(
            resource_type,
            dest,
            resource.name
        ))


@cli.group('orders')
def orders():
    '''Commands for interacting with the Orders API'''
    pass


@orders.command('list')
# @click.option('--status', help="'all', 'in-progress', 'completed'")
@pretty
def list_orders(pretty):
    '''List all pending order requests; optionally filter by status'''
    cl = clientv1()
    echo_json_response(call_and_wrap(cl.get_orders), pretty)


@orders.command('get')
@click.argument('order_id', type=click.UUID)
@pretty
def get_order(order_id, pretty):
    '''Get order request for a given order ID'''
    cl = clientv1()
    echo_json_response(call_and_wrap(cl.get_individual_order, order_id),
                       pretty)


@orders.command('cancel')
@click.argument('order_id', type=click.UUID)
@pretty
def cancel_order(order_id, pretty):
    '''Cancel a running order by given order ID'''
    cl = clientv1()
    echo_json_response(call_and_wrap(cl.cancel_order, order_id), pretty)


@click.option('--name', required=True)
@click.option('--id', help='One or more comma-separated item IDs',
              cls=RequiredUnless, this_opt_exists='ids_from_search')
# Note: This is passed as a string, because --item-type is a required field for
# both 'data search' and 'orders create'.
@click.option('--ids_from_search',
              help='Embedded data search')
@click.option('--clip', type=ClipAOI(),
              help='Provide a GeoJSON AOI Geometry for clipping')
@click.option('--email', default=False, is_flag=True,
              help='Send email notification when Order is complete')
@click.option('--zip', type=click.Choice(['order', 'bundle']),
              help='Receive output of toolchain as a .zip archive.')
@click.option('--cloudconfig', help=('Path to cloud delivery config'),
              type=click.Path(exists=True, resolve_path=True, readable=True,
                              allow_dash=False, dir_okay=False,
                              file_okay=True))
@click.option('--tools', help=('Path to toolchain json'),
              type=click.Path(exists=True, resolve_path=True, readable=True,
                              allow_dash=False, dir_okay=False,
                              file_okay=True))
@bundle_option
@click.option(
    '--item-type', multiple=False, required=True, type=ItemType(), help=(
        'Specify an item type'
    )
)
@orders.command('create')
@pretty
def create_order(pretty, **kwargs):
    '''Create an order'''
    ids_from_search = kwargs.get('ids_from_search')
    if ids_from_search is not None:
        runner = CliRunner()
        resp = runner.invoke(quick_search, ids_from_search).output
        try:
            id_list = ids_from_search_response(resp)
        except ValueError:
            raise click.ClickException('ids_from_search, {}'.format(resp))
        kwargs['id'] = id_list
        del kwargs['ids_from_search']
    cl = clientv1()
    request = create_order_request(**kwargs)
    echo_json_response(call_and_wrap(cl.create_order, request), pretty)


@orders.command('download')
@click.argument('order_id', type=click.UUID)
@click.option('--quiet', is_flag=True, help=(
    'Disable ANSI control output'
))
@click.option('--dest', default='.', help=(
    'Location to download files to'), type=click.Path(
        exists=True, resolve_path=True, writable=True, file_okay=False
))
@pretty
def download_order(order_id, dest, quiet, pretty):
    '''Download an order by given order ID'''
    cl = clientv1()
    dl = downloader.create(cl, order=True)

    output = downloader_output(dl, disable_ansi=quiet)
    output.start()

    items = cl.get_individual_order(order_id).items_iter(limit=None)
    handle_interrupt(dl.shutdown, dl.download, items, [], dest)
